from math import ceil
from math import pi as kPI
from typing import Optional
from typing import Tuple
from typing import Union

import torch
import torch.nn.functional as fn
from torch import Tensor
from torch.distributions.normal import Normal
from torch.distributions.utils import broadcast_all

import pfhedge.autogreek as autogreek
from pfhedge._utils.typing import TensorOrScalar


def european_payoff(input: Tensor, call: bool = True, strike: float = 1.0) -> Tensor:
    """Returns the payoff of a European option.

    .. seealso::
        - :class:`pfhedge.instruments.EuropeanOption`

    Args:
        input (torch.Tensor): The input tensor representing the price trajectory.
        call (bool, default=True): Specifies whether the option is call or put.
        strike (float, default=1.0): The strike price of the option.

    Shape:
        - input: :math:`(*, T)` where
          :math:`T` is the number of time steps and
          :math:`*` means any number of additional dimensions.
        - output: :math:`(*)`

    Returns:
        torch.Tensor
    """
    if call:
        return fn.relu(input[..., -1] - strike)
    else:
        return fn.relu(strike - input[..., -1])


def lookback_payoff(input: Tensor, call: bool = True, strike: float = 1.0) -> Tensor:
    """Returns the payoff of a lookback option with a fixed strike.

    .. seealso::
        - :class:`pfhedge.instruments.LookbackOption`

    Args:
        input (torch.Tensor): The input tensor representing the price trajectory.
        call (bool, default=True): Specifies whether the option is call or put.
        strike (float, default=1.0): The strike price of the option.

    Shape:
        - input: :math:`(*, T)` where
          :math:`T` is the number of time steps and
          :math:`*` means any number of additional dimensions.
        - output: :math:`(*)`

    Returns:
        torch.Tensor
    """
    if call:
        return fn.relu(input.max(dim=-1).values - strike)
    else:
        return fn.relu(strike - input.min(dim=-1).values)


def american_binary_payoff(
    input: Tensor, call: bool = True, strike: float = 1.0
) -> Tensor:
    """Returns the payoff of an American binary option.

    .. seealso::
        - :class:`pfhedge.instruments.AmericanBinaryOption`

    Args:
        input (torch.Tensor): The input tensor representing the price trajectory.
        call (bool, default=True): Specifies whether the option is call or put.
        strike (float, default=1.0): The strike price of the option.

    Shape:
        - input: :math:`(*, T)` where
          :math:`T` is the number of time steps and
          :math:`*` means any number of additional dimensions.
        - output: :math:`(*)`

    Returns:
        torch.Tensor
    """
    if call:
        return (input.max(dim=-1).values >= strike).to(input)
    else:
        return (input.min(dim=-1).values <= strike).to(input)


def european_binary_payoff(
    input: Tensor, call: bool = True, strike: float = 1.0
) -> Tensor:
    """Returns the payoff of a European binary option.

    .. seealso::
        - :class:`pfhedge.instruments.EuropeanBinaryOption`

    Args:
        input (torch.Tensor): The input tensor representing the price trajectory.
        call (bool, default=True): Specifies whether the option is call or put.
        strike (float, default=1.0): The strike price of the option.

    Shape:
        - input: :math:`(*, T)` where
          :math:`T` is the number of time steps and
          :math:`*` means any number of additional dimensions.
        - output: :math:`(*)`

    Returns:
        torch.Tensor
    """
    if call:
        return (input[..., -1] >= strike).to(input)
    else:
        return (input[..., -1] <= strike).to(input)


def european_forward_start_payoff(
    input: Tensor, strike: float = 1.0, start_index: int = 0, end_index: int = -1
) -> Tensor:
    """Returns the payoff of a European forward start option.

    .. seealso::
        - :class:`pfhedge.instruments.EuropeanForwardStartOption`

    Args:
        input (torch.Tensor): The input tensor representing the price trajectory.
        strike (float, default=1.0): The strike price of the option.
        start_index (int, default=0): The time index at which the option starts.
        end_index (int, default=-1): The time index at which the option ends.

    Shape:
        - input: :math:`(*, T)` where
          :math:`T` is the number of time steps and
          :math:`*` means any number of additional dimensions.
        - output: :math:`(*)`

    Returns:
        torch.Tensor
    """
    return fn.relu(input[..., end_index] / input[..., start_index] - strike)


def exp_utility(input: Tensor, a: float = 1.0) -> Tensor:
    r"""Applies an exponential utility function.

    An exponential utility function is defined as:

    .. math::

        u(x) = -\exp(-a x) \,.

    Args:
        input (torch.Tensor): The input tensor.
        a (float, default=1.0): The risk aversion coefficient of the exponential
            utility.

    Returns:
        torch.Tensor
    """
    return -(-a * input).exp()


def isoelastic_utility(input: Tensor, a: float) -> Tensor:
    r"""Applies an isoelastic utility function.

    An isoelastic utility function is defined as:

    .. math::

        u(x) = \begin{cases}
        x^{1 - a} & a \neq 1 \\
        \log{x} & a = 1
        \end{cases} \,.

    Args:
        input (torch.Tensor): The input tensor.
        a (float): Relative risk aversion coefficient of the isoelastic
            utility.

    Returns:
        torch.Tensor
    """
    if a == 1.0:
        return input.log()
    else:
        return input.pow(1.0 - a)


def entropic_risk_measure(input: Tensor, a: float = 1.0) -> Tensor:
    """Returns the entropic risk measure.

    See :class:`pfhedge.nn.EntropicRiskMeasure` for details.
    """
    return (-exp_utility(input, a=a).mean(0)).log() / a


def topp(input: Tensor, p: float, dim: Optional[int] = None, largest: bool = True):
    """Returns the largest :math:`p * N` elements of the given input tensor,
    where :math:`N` stands for the total number of elements in the input tensor.

    If ``dim`` is not given, the last dimension of the ``input`` is chosen.

    If ``largest`` is ``False`` then the smallest elements are returned.

    A namedtuple of ``(values, indices)`` is returned, where the ``indices``
    are the indices of the elements in the original ``input`` tensor.

    .. seealso::
        - :func:`torch.topk`: Returns the ``k`` largest elements of the given input tensor
          along a given dimension.

    Args:
        input (torch.Tensor): The input tensor.
        p (float): The quantile level.
        dim (int, optional): The dimension to sort along.
        largest (bool, default=True): Controls whether to return largest or smallest
            elements.

    Returns:
        torch.Tensor

    Examples:
        >>> from pfhedge.nn.functional import topp
        >>>
        >>> input = torch.arange(1.0, 6.0)
        >>> input
        tensor([1., 2., 3., 4., 5.])
        >>> topp(input, 3 / 5)
        torch.return_types.topk(
        values=tensor([5., 4., 3.]),
        indices=tensor([4, 3, 2]))
    """
    if dim is None:
        return input.topk(ceil(p * input.numel()), largest=largest)
    else:
        return input.topk(ceil(p * input.size(dim)), dim=dim, largest=largest)


def expected_shortfall(input: Tensor, p: float, dim: Optional[int] = None) -> Tensor:
    """Returns the expected shortfall of the given input tensor.

    Args:
        input (torch.Tensor): The input tensor.
        p (float): The quantile level.
        dim (int, optional): The dimension to sort along.

    Returns:
        torch.Tensor

    Examples:
        >>> from pfhedge.nn.functional import expected_shortfall
        >>>
        >>> input = -torch.arange(10.0)
        >>> input
        tensor([-0., -1., -2., -3., -4., -5., -6., -7., -8., -9.])
        >>> expected_shortfall(input, 0.3)
        tensor(8.)
    """
    if dim is None:
        return -topp(input, p=p, largest=False).values.mean()
    else:
        return -topp(input, p=p, largest=False, dim=dim).values.mean(dim=dim)


def _min_values(input: Tensor, dim: Optional[int] = None) -> Tensor:
    return input.min() if dim is None else input.min(dim=dim).values


def _max_values(input: Tensor, dim: Optional[int] = None) -> Tensor:
    return input.max() if dim is None else input.max(dim=dim).values


def value_at_risk(input: Tensor, p: float, dim: Optional[int] = None) -> Tensor:
    """Returns the value at risk of the given input tensor.

    Note:
        If :math:`p \leq 1 / N`` with :math:`N` being the number of elements to sort,
        returns the smallest element in the tensor.
        If :math:`p > 1 - 1 / N``, returns the largest element in the tensor.

    Args:
        input (torch.Tensor): The input tensor.
        p (float): The quantile level.
        dim (int, optional): The dimension to sort along.

    Returns:
        torch.Tensor

    Examples:
        >>> from pfhedge.nn.functional import value_at_risk
        >>>
        >>> input = -torch.arange(10.0)
        >>> input
        tensor([-0., -1., -2., -3., -4., -5., -6., -7., -8., -9.])
        >>> value_at_risk(input, 0.3)
        tensor(-7.)
    """
    n = input.numel() if dim is None else input.size(dim)

    if p <= 1 / n:
        output = _min_values(input, dim=dim)
    elif p > 1 - 1 / n:
        output = _max_values(input, dim=dim)
    else:
        q = (p - (1 / n)) / (1 - (1 / n))
        output = input.quantile(q, dim=dim)

    return output


def leaky_clamp(
    input: Tensor,
    min: Optional[Tensor] = None,
    max: Optional[Tensor] = None,
    clamped_slope: float = 0.01,
    inverted_output: str = "mean",
) -> Tensor:
    r"""Leakily clamp all elements in ``input`` into the range :math:`[\min, \max]`.

    See :class:`pfhedge.nn.LeakyClamp` for details.
    """
    x = input

    if min is not None:
        min = torch.as_tensor(min).to(x)
        x = x.maximum(min + clamped_slope * (x - min))

    if max is not None:
        max = torch.as_tensor(max).to(x)
        x = x.minimum(max + clamped_slope * (x - max))

    if min is not None and max is not None:
        if inverted_output == "mean":
            y = (min + max) / 2
        elif inverted_output == "max":
            y = max
        else:
            raise ValueError("inverted_output must be 'mean' or 'max'.")
        x = x.where(min <= max, y)

    return x


def clamp(
    input: Tensor,
    min: Optional[Tensor] = None,
    max: Optional[Tensor] = None,
    inverted_output: str = "mean",
) -> Tensor:
    r"""Clamp all elements in ``input`` into the range :math:`[\min, \max]`.

    See :class:`pfhedge.nn.Clamp` for details.
    """
    if inverted_output == "mean":
        output = leaky_clamp(input, min, max, clamped_slope=0.0, inverted_output="mean")
    elif inverted_output == "max":
        output = torch.clamp(input, min, max)
    else:
        raise ValueError("inverted_output must be 'mean' or 'max'.")
    return output


def realized_variance(input: Tensor, dt: TensorOrScalar) -> Tensor:
    r"""Returns the realized variance of the price.

    Realized variance :math:`\sigma^2` of the stock price :math:`S` is defined as:

    .. math::

        \sigma^2 = \frac{1}{T - 1} \sum_{i = 1}^{T - 1}
        \frac{1}{dt} \log(S_{i + 1} / S_i)^2

    where :math:`T` is the number of time steps.

    Note:
        The mean of log return is assumed to be zero.

    Args:
        input (torch.Tensor): The input tensor.
        dt (torch.Tensor or float): The intervals of the time steps.

    Shape:
        - input: :math:`(*, T)` where
          :math:`T` stands for the number of time steps and
          :math:`*` means any number of additional dimensions.
        - output: :math:`(*)`

    Returns:
        torch.Tensor
    """
    return input.log().diff(dim=-1).square().mean(dim=-1) / dt


def realized_volatility(input: Tensor, dt: Union[Tensor, float]) -> Tensor:
    """Returns the realized volatility of the price.
    It is square root of :func:`realized_variance`.

    Args:
        input (torch.Tensor): The input tensor.
        dt (torch.Tensor or float): The intervals of the time steps.

    Shape:
        - input: :math:`(*, T)` where
          :math:`T` stands for the number of time steps and
          :math:`*` means any number of additional dimensions.
        - output: :math:`(*)`

    Returns:
        torch.Tensor
    """
    return realized_variance(input, dt=dt).sqrt()


def terminal_value(
    spot: Tensor,
    unit: Tensor,
    cost: float = 0.0,
    payoff: Optional[Tensor] = None,
    deduct_first_cost: bool = True,
) -> Tensor:
    r"""Returns the terminal portfolio value.

    The terminal value of a hedger's portfolio is given by

    .. math::

        \text{PL}(Z, \delta, S) =
        - Z
        + \sum_{i = 0}^{T - 2} \delta_{i - 1} (S_{i} - S_{i - 1})
        - c \sum_{i = 0}^{T - 1} |\delta_{i} - \delta_{i - 1}| S_{i}

    where :math:`Z` is the payoff of the derivative, :math:`T` is the number of
    time steps, :math:`S` is the spot price, :math:`\delta` is the signed number
    of shares held at each time step.
    We define :math:`\delta_0 = 0` for notational convenience.

    A hedger sells the derivative to its customer and
    obliges to settle the payoff at maturity.
    The dealer hedges the risk of this liability
    by trading the underlying instrument of the derivative.
    The resulting profit and loss is obtained by adding up the payoff to the
    customer, capital gains from the underlying asset, and the transaction cost.

    References:
        - Buehler, H., Gonon, L., Teichmann, J. and Wood, B., 2019.
          Deep hedging. Quantitative Finance, 19(8), pp.1271-1291.
          [arXiv:`1802.03042 <https://arxiv.org/abs/1802.03042>`_ [q-fin]]

    Args:
        spot (torch.Tensor): The spot price of the underlying asset :math:`S`.
        unit (torch.Tensor): The signed number of shares of the underlying asset
            :math:`\delta`.
        cost (float, default=0.0): The proportional transaction cost rate of
            the underlying asset :math:`c`.
        payoff (torch.Tensor, optional): The payoff of the derivative :math:`Z`.
        deduct_first_cost (bool, default=True): Whether to deduct the transaction
            cost of the stock at the first time step.
            If ``False``, :math:`- c |\delta_0| S_1` is omitted the above
            equation of the terminal value.

    Shape:
        - spot: :math:`(N, *, T)` where
          :math:`T` is the number of time steps and
          :math:`*` means any number of additional dimensions.
        - unit: :math:`(N, *, T)`
        - payoff: :math:`(N, *)`
        - output: :math:`(N, *)`.

    Returns:
        torch.Tensor
    """
    if spot.size() != unit.size():
        raise RuntimeError(f"unmatched sizes: spot {spot.size()}, unit {unit.size()}")
    if payoff is not None and spot.size()[:-1] != payoff.size():
        raise RuntimeError(
            f"unmatched sizes: spot {spot.size()}, payoff {payoff.size()}"
        )

    value = unit[..., :-1].mul(spot.diff(dim=-1)).sum(-1)
    value += -cost * unit.diff(dim=-1).abs().mul(spot[..., 1:]).sum(-1)
    if payoff is not None:
        value -= payoff
    if deduct_first_cost:
        value -= cost * unit[..., 0].abs() * spot[..., 0]

    return value


def ncdf(input: Tensor) -> Tensor:
    r"""Returns a new tensor with the normal cumulative distribution function.

    .. math::
        \text{ncdf}(x) =
            \int_{-\infty}^x
            \frac{1}{\sqrt{2 \pi}} e^{-\frac{y^2}{2}} dy

    Args:
        input (torch.Tensor): The input tensor.

    Returns:
        torch.Tensor

    Examples:
        >>> from pfhedge.nn.functional import ncdf
        >>>
        >>> input = torch.tensor([-1.0, 0.0, 10.0])
        >>> ncdf(input)
        tensor([0.1587, 0.5000, 1.0000])
    """
    return Normal(0.0, 1.0).cdf(input)


def npdf(input: Tensor) -> Tensor:
    r"""Returns a new tensor with the normal distribution function.

    .. math::
        \text{npdf}(x) = \frac{1}{\sqrt{2 \pi}} e^{-\frac{x^2}{2}}

    Args:
        input (torch.Tensor): The input tensor.

    Returns:
        torch.Tensor

    Examples:
        >>> from pfhedge.nn.functional import npdf
        >>>
        >>> input = torch.tensor([-1.0, 0.0, 10.0])
        >>> npdf(input)
        tensor([2.4197e-01, 3.9894e-01, 7.6946e-23])
    """
    return Normal(0.0, 1.0).log_prob(input).exp()


def d1(
    log_moneyness: TensorOrScalar,
    time_to_maturity: TensorOrScalar,
    volatility: TensorOrScalar,
) -> Tensor:
    r"""Returns :math:`d_1` in the Black-Scholes formula.

    .. math::
        d_1 = \frac{s}{\sigma \sqrt{t}} + \frac{\sigma \sqrt{t}}{2}

    where
    :math:`s` is the log moneyness,
    :math:`t` is the time to maturity, and
    :math:`\sigma` is the volatility.

    Note:
        Risk-free rate is set to zero.

    Args:
        log_moneyness (torch.Tensor or float): Log moneyness of the underlying asset.
        time_to_maturity (torch.Tensor or float): Time to maturity of the derivative.
        volatility (torch.Tensor or float): Volatility of the underlying asset.

    Returns:
        torch.Tensor
    """
    s, t, v = broadcast_all(log_moneyness, time_to_maturity, volatility)
    if not (t >= 0).all():
        raise ValueError("all elements in time_to_maturity have to be non-negative")
    if not (v >= 0).all():
        raise ValueError("all elements in volatility have to be non-negative")
    variance = v * t.sqrt()
    output = s / variance + variance / 2
    # TODO(simaki): Replace zeros_like with 0.0 once https://github.com/pytorch/pytorch/pull/62084 is merged
    return output.where((s != 0).logical_or(variance != 0), torch.zeros_like(output))


def d2(
    log_moneyness: TensorOrScalar,
    time_to_maturity: TensorOrScalar,
    volatility: TensorOrScalar,
) -> Tensor:
    r"""Returns :math:`d_2` in the Black-Scholes formula.

    .. math::
        d_2 = \frac{s}{\sigma \sqrt{t}} - \frac{\sigma \sqrt{t}}{2}

    where
    :math:`s` is the log moneyness,
    :math:`t` is the time to maturity, and
    :math:`\sigma` is the volatility.

    Note:
        Risk-free rate is set to zero.

    Args:
        log_moneyness (torch.Tensor or float): Log moneyness of the underlying asset.
        time_to_maturity (torch.Tensor or float): Time to maturity of the derivative.
        volatility (torch.Tensor or float): Volatility of the underlying asset.

    Returns:
        torch.Tensor
    """
    s, t, v = broadcast_all(log_moneyness, time_to_maturity, volatility)
    if not (t >= 0).all():
        raise ValueError("all elements in time_to_maturity have to be non-negative")
    if not (v >= 0).all():
        raise ValueError("all elements in volatility have to be non-negative")
    variance = v * t.sqrt()
    output = s / variance - variance / 2
    # TODO(simaki): Replace zeros_like with 0.0 once https://github.com/pytorch/pytorch/pull/62084 is merged
    return output.where((s != 0).logical_or(variance != 0), torch.zeros_like(output))


def ww_width(
    gamma: Tensor, spot: Tensor, cost: TensorOrScalar, a: TensorOrScalar = 1.0
) -> Tensor:
    r"""Returns half-width of the no-transaction band for
    Whalley-Wilmott's hedging strategy.

    See :class:`pfhedge.nn.WhalleyWilmott` for details.

    Args:
        gamma (torch.Tensor): The gamma of the derivative,
        spot (torch.Tensor): The spot price of the underlier.
        cost (torch.Tensor or float): The cost rate of the underlier.
        a (torch.Tensor or float, default=1.0): Risk aversion parameter in exponential utility.

    Returns:
        torch.Tensor
    """
    return (cost * (3 / 2) * gamma.square() * spot / a).pow(1 / 3)


def svi_variance(
    input: TensorOrScalar,
    a: TensorOrScalar,
    b: TensorOrScalar,
    rho: TensorOrScalar,
    m: TensorOrScalar,
    sigma: TensorOrScalar,
) -> Tensor:
    r"""Returns variance in the SVI model.

    See :class:`pfhedge.nn.SVIVariance` for details.

    Args:
        input (torch.Tensor or float): Log strike of the underlying asset.
            That is, :math:`k = \log(K / S)` for spot :math:`S` and strike :math:`K`.
        a (torch.Tensor or float): The parameter :math:`a`.
        b (torch.Tensor or float): The parameter :math:`b`.
        rho (torch.Tensor or float): The parameter :math:`\rho`.
        m (torch.Tensor or float): The parameter :math:`m`.
        sigma (torch.Tensor or float): The parameter :math:`s`.

    Returns:
        torch.Tensor
    """
    k_m = torch.as_tensor(input - m)  # k - m
    return a + b * (rho * k_m + (k_m.square() + sigma ** 2).sqrt())


def bilerp(
    input1: Tensor,
    input2: Tensor,
    input3: Tensor,
    input4: Tensor,
    weight1: TensorOrScalar,
    weight2: TensorOrScalar,
) -> Tensor:
    r"""Does a bilinear interpolation of four tensors based on a scalar or tensor weights and
    returns the resulting tensor.

    The output is given by

    .. math::
        \text{output}_i
        & = (1 - w_1) (1 - w_2) \cdot \text{input1}_i
        + w_1 (1 - w_2) \cdot \text{input2}_i \\
        & \quad + (1 - w_1) w_2 \cdot \text{input3}_i
        + w_1 w_2 \cdot \text{input4}_i ,

    where :math:`w_1` and :math:`w_2` are the weights.

    The shapes of inputs must be broadcastable.
    If ``weight`` is a tensor, then the shapes of ``weight`` must also be broadcastable.

    Args:
        input1 (torch.Tensor): The input tensor.
        input2 (torch.Tensor): The input tensor.
        input3 (torch.Tensor): The input tensor.
        input4 (torch.Tensor): The input tensor.
        weight1 (float or torch.Tensor): The weight tensor.
        weight2 (float or torch.Tensor): The weight tensor.

    Returns:
        torch.Tensor
    """
    lerp1 = torch.lerp(input1, input2, weight1)
    lerp2 = torch.lerp(input3, input4, weight1)
    return torch.lerp(lerp1, lerp2, weight2)


def bs_european_price(
    log_moneyness: Tensor,
    time_to_maturity: Tensor,
    volatility: Tensor,
    strike: TensorOrScalar = 1.0,
    call: bool = True,
) -> Tensor:
    """Returns Black-Scholes price of a European option.

    See :func:`pfhedge.nn.BSEuropeanOption.price` for details.
    """
    s, t, v = broadcast_all(log_moneyness, time_to_maturity, volatility)

    price = strike * (s.exp() * ncdf(d1(s, t, v)) - ncdf(d2(s, t, v)))
    price = price + strike * (1 - s.exp()) if not call else price  # put-call parity

    return price


def bs_european_delta(
    log_moneyness: Tensor,
    time_to_maturity: Tensor,
    volatility: Tensor,
    call: bool = True,
) -> Tensor:
    """Returns Black-Scholes delta of a European option.

    See :func:`pfhedge.nn.BSEuropeanOption.delta` for details.
    """
    s, t, v = broadcast_all(log_moneyness, time_to_maturity, volatility)

    delta = ncdf(d1(s, t, v))
    delta = delta - 1 if not call else delta  # put-call parity

    return delta


def bs_european_gamma(
    log_moneyness: Tensor,
    time_to_maturity: Tensor,
    volatility: Tensor,
    strike: TensorOrScalar = 1.0,
) -> Tensor:
    """Returns Black-Scholes gamma of a European option.

    See :func:`pfhedge.nn.BSEuropeanOption.gamma` for details.
    """
    s, t, v = broadcast_all(log_moneyness, time_to_maturity, volatility)
    spot = strike * s.exp()
    numerator = npdf(d1(s, t, v))
    denominator = spot * v * t.sqrt()
    output = numerator / denominator
    return torch.where(
        (numerator == 0).logical_and(denominator == 0), torch.zeros_like(output), output
    )


def bs_european_vega(
    log_moneyness: Tensor,
    time_to_maturity: Tensor,
    volatility: Tensor,
    strike: TensorOrScalar,
) -> Tensor:
    """Returns Black-Scholes vega of a European option.

    See :func:`pfhedge.nn.BSEuropeanOption.vega` for details.
    """
    s, t, v = broadcast_all(log_moneyness, time_to_maturity, volatility)
    price = strike * s.exp()
    return npdf(d1(s, t, v)) * price * t.sqrt()


def bs_european_theta(
    log_moneyness: Tensor,
    time_to_maturity: Tensor,
    volatility: Tensor,
    strike: TensorOrScalar,
) -> Tensor:
    """Returns Black-Scholes theta of a European option.

    See :func:`pfhedge.nn.BSEuropeanOption.theta` for details.
    """
    s, t, v = broadcast_all(log_moneyness, time_to_maturity, volatility)
    price = strike * s.exp()
    numerator = -npdf(d1(s, t, v)) * price * v
    denominator = 2 * t.sqrt()
    output = numerator / denominator
    return torch.where(
        (numerator == 0).logical_and(denominator == 0), torch.zeros_like(output), output
    )


def bs_european_binary_price(
    log_moneyness: Tensor,
    time_to_maturity: Tensor,
    volatility: Tensor,
    call: bool = True,
) -> Tensor:
    """Returns Black-Scholes price of a European binary option.

    See :func:`pfhedge.nn.BSEuropeanBinaryOption.price` for details.
    """
    s, t, v = broadcast_all(log_moneyness, time_to_maturity, volatility)

    price = ncdf(d2(s, t, v))
    price = 1.0 - price if not call else price  # put-call parity

    return price


def bs_european_binary_delta(
    log_moneyness: Tensor,
    time_to_maturity: Tensor,
    volatility: Tensor,
    call: bool = True,
    strike: TensorOrScalar = 1.0,
) -> Tensor:
    """Returns Black-Scholes delta of a European binary option.

    See :func:`pfhedge.nn.BSEuropeanBinaryOption.delta` for details.
    """
    s, t, v = broadcast_all(log_moneyness, time_to_maturity, volatility)

    spot = s.exp() * strike

    numerator = npdf(d2(s, t, v))
    denominator = spot * v * t.sqrt()
    delta = numerator / denominator
    delta = torch.where(
        (numerator == 0).logical_and(denominator == 0), torch.zeros_like(delta), delta
    )
    delta = -delta if not call else delta  # put-call parity

    return delta


def bs_european_binary_gamma(
    log_moneyness: Tensor,
    time_to_maturity: Tensor,
    volatility: Tensor,
    strike: TensorOrScalar = 1.0,
) -> Tensor:
    """Returns Black-Scholes gamma of a European binary option.

    See :func:`pfhedge.nn.BSEuropeanBinaryOption.gamma` for details.
    """
    # TODO(simaki): Directly compute gamma.
    return autogreek.gamma(
        bs_european_binary_price,
        log_moneyness=log_moneyness,
        time_to_maturity=time_to_maturity,
        volatility=volatility,
        strike=strike,
    )


def bs_european_binary_vega(
    log_moneyness: Tensor,
    time_to_maturity: Tensor,
    volatility: Tensor,
    strike: TensorOrScalar = 1.0,
) -> Tensor:
    """Returns Black-Scholes vega of a European binary option.

    See :func:`pfhedge.nn.BSEuropeanBinaryOption.vega` for details.
    """
    # TODO(simaki): Directly compute gamma.
    return autogreek.vega(
        bs_european_binary_price,
        log_moneyness=log_moneyness,
        time_to_maturity=time_to_maturity,
        volatility=volatility,
        strike=strike,
    )


def bs_european_binary_theta(
    log_moneyness: Tensor,
    time_to_maturity: Tensor,
    volatility: Tensor,
    strike: TensorOrScalar = 1.0,
) -> Tensor:
    """Returns Black-Scholes theta of a European binary option.

    See :func:`pfhedge.nn.BSEuropeanBinaryOption.theta` for details.
    """
    # TODO(simaki): Directly compute theta.
    return autogreek.theta(
        bs_european_binary_price,
        log_moneyness=log_moneyness,
        time_to_maturity=time_to_maturity,
        volatility=volatility,
        strike=strike,
    )


def bs_american_binary_price(
    log_moneyness: Tensor,
    max_log_moneyness: Tensor,
    time_to_maturity: Tensor,
    volatility: Tensor,
) -> Tensor:
    """Returns Black-Scholes price of an American binary option.

    See :func:`pfhedge.nn.BSAmericanBinaryOption.price` for details.
    """
    # This formula is derived using the results in Section 7.3.3 of Shreve's book.
    # Price is I_2 - I_4 where the interval of integration is [k --> -inf, b].
    # By this substitution we get N([log(S(0) / K) + ...] / sigma T) --> 1.

    s, t, v = broadcast_all(log_moneyness, time_to_maturity, volatility)
    p = ncdf(d2(s, t, v)) + s.exp() * (1 - ncdf(d2(-s, t, v)))

    return p.where(max_log_moneyness < 0, torch.ones_like(p))


def bs_american_binary_delta(
    log_moneyness: Tensor,
    max_log_moneyness: Tensor,
    time_to_maturity: Tensor,
    volatility: Tensor,
    strike: TensorOrScalar,
) -> Tensor:
    """Returns Black-Scholes delta of an American binary option.

    See :func:`pfhedge.nn.BSAmericanBinaryOption.delta` for details.
    """
    s, t, v = broadcast_all(log_moneyness, time_to_maturity, volatility)
    spot = s.exp() * strike
    # ToDo: fix 0/0 issue
    p = (
        npdf(d2(s, t, v)) / (spot * v * t.sqrt())
        - (1 - ncdf(d2(-s, t, v))) / strike
        + npdf(d2(-s, t, v)) / (strike * v * t.sqrt())
    )

    return p.where(max_log_moneyness < 0, torch.zeros_like(p))


def bs_american_binary_gamma(
    log_moneyness: Tensor,
    max_log_moneyness: Tensor,
    time_to_maturity: Tensor,
    volatility: Tensor,
    strike: TensorOrScalar,
) -> Tensor:
    """Returns Black-Scholes gamma of an American binary option.

    See :func:`pfhedge.nn.BSAmericanBinaryOption.gamma` for details.
    """
    # TODO(simaki): Compute analytically
    return autogreek.gamma(
        bs_american_binary_price,
        log_moneyness=log_moneyness,
        max_log_moneyness=max_log_moneyness,
        time_to_maturity=time_to_maturity,
        volatility=volatility,
        strike=strike,
    )


def bs_american_binary_vega(
    log_moneyness: Tensor,
    max_log_moneyness: Tensor,
    time_to_maturity: Tensor,
    volatility: Tensor,
    strike: TensorOrScalar,
) -> Tensor:
    """Returns Black-Scholes vega of an American binary option.

    See :func:`pfhedge.nn.BSAmericanBinaryOption.vega` for details.
    """
    # TODO(simaki): Compute analytically
    return autogreek.vega(
        bs_american_binary_price,
        log_moneyness=log_moneyness,
        max_log_moneyness=max_log_moneyness,
        time_to_maturity=time_to_maturity,
        volatility=volatility,
        strike=strike,
    )


def bs_american_binary_theta(
    log_moneyness: Tensor,
    max_log_moneyness: Tensor,
    time_to_maturity: Tensor,
    volatility: Tensor,
    strike: TensorOrScalar,
) -> Tensor:
    """Returns Black-Scholes theta of an American binary option.

    See :func:`pfhedge.nn.BSAmericanBinaryOption.theta` for details.
    """
    # TODO(simaki): Compute analytically
    return autogreek.theta(
        bs_american_binary_price,
        log_moneyness=log_moneyness,
        max_log_moneyness=max_log_moneyness,
        time_to_maturity=time_to_maturity,
        volatility=volatility,
        strike=strike,
    )


def bs_lookback_price(
    log_moneyness: Tensor,
    max_log_moneyness: Tensor,
    time_to_maturity: Tensor,
    volatility: Tensor,
    strike: TensorOrScalar,
) -> Tensor:
    """Returns Black-Scholes price of a lookback option.

    See :func:`pfhedge.nn.BSLookbackOption.price` for details.
    """
    s, m, t, v = map(
        torch.as_tensor,
        (log_moneyness, max_log_moneyness, time_to_maturity, volatility),
    )

    spot = s.exp() * strike
    max = m.exp() * strike
    d1_value = d1(s, t, v)
    d2_value = d2(s, t, v)
    m1 = d1(s - m, t, v)  # d' in the paper
    m2 = d2(s - m, t, v)

    # when max < strike
    price_0 = spot * (
        ncdf(d1_value) + v * t.sqrt() * (d1_value * ncdf(d1_value) + npdf(d1_value))
    ) - strike * ncdf(d2_value)
    # when max >= strike
    price_1 = (
        spot * (ncdf(m1) + v * t.sqrt() * (m1 * ncdf(m1) + npdf(m1)))
        - strike
        + max * (1 - ncdf(m2))
    )

    return torch.where(max < strike, price_0, price_1)


def bs_lookback_delta(
    log_moneyness: Tensor,
    max_log_moneyness: Tensor,
    time_to_maturity: Tensor,
    volatility: Tensor,
    strike: TensorOrScalar,
) -> Tensor:
    """Returns Black-Scholes delta of a lookback option.

    See :func:`pfhedge.nn.BSLookbackOption.delta` for details.
    """
    # TODO(simaki): Calculate analytically
    return autogreek.delta(
        bs_lookback_price,
        log_moneyness=log_moneyness,
        max_log_moneyness=max_log_moneyness,
        time_to_maturity=time_to_maturity,
        volatility=volatility,
        strike=strike,
    )


def bs_lookback_gamma(
    log_moneyness: Tensor,
    max_log_moneyness: Tensor,
    time_to_maturity: Tensor,
    volatility: Tensor,
    strike: TensorOrScalar,
) -> Tensor:
    """Returns Black-Scholes gamma of a lookback option.

    See :func:`pfhedge.nn.BSLookbackOption.gamma` for details.
    """
    # TODO(simaki): Calculate analytically
    return autogreek.gamma(
        bs_lookback_price,
        log_moneyness=log_moneyness,
        max_log_moneyness=max_log_moneyness,
        time_to_maturity=time_to_maturity,
        volatility=volatility,
        strike=strike,
    )


def bs_lookback_vega(
    log_moneyness: Tensor,
    max_log_moneyness: Tensor,
    time_to_maturity: Tensor,
    volatility: Tensor,
    strike: TensorOrScalar,
) -> Tensor:
    """Returns Black-Scholes vega of a lookback option.

    See :func:`pfhedge.nn.BSLookbackOption.vega` for details.
    """
    # TODO(simaki): Calculate analytically
    return autogreek.vega(
        bs_lookback_price,
        log_moneyness=log_moneyness,
        max_log_moneyness=max_log_moneyness,
        time_to_maturity=time_to_maturity,
        volatility=volatility,
        strike=strike,
    )


def bs_lookback_theta(
    log_moneyness: Tensor,
    max_log_moneyness: Tensor,
    time_to_maturity: Tensor,
    volatility: Tensor,
    strike: TensorOrScalar,
) -> Tensor:
    """Returns Black-Scholes theta of a lookback option.

    See :func:`pfhedge.nn.BSLookbackOption.theta` for details.
    """
    # TODO(simaki): Calculate analytically
    return autogreek.theta(
        bs_lookback_price,
        log_moneyness=log_moneyness,
        max_log_moneyness=max_log_moneyness,
        time_to_maturity=time_to_maturity,
        volatility=volatility,
        strike=strike,
    )


def box_muller(
    input1: Tensor, input2: Tensor, epsilon: float = 1e-10
) -> Tuple[Tensor, Tensor]:
    r"""Returns two tensors obtained by applying Box-Muller transformation to two input tensors.

    .. math::
        & \mathrm{output1}_i
            = \sqrt{- 2 \log (\mathrm{input1}_i)} \cos(2 \pi \cdot \mathrm{input2}_i) , \\
        & \mathrm{output2}_i
            = \sqrt{- 2 \log (\mathrm{input1}_i)} \sin(2 \pi \cdot \mathrm{input2}_i) .

    Args:
        input1 (torch.Tensor): The first input tensor.
        input2 (torch.Tensor): The second input tensor.
        epsilon (float, default=1e-10): A small constant to avoid evaluating :math:`\log(0)`.
            The tensor ``input1`` will be clamped with this value being the minimum.

    Returns:
        (torch.Tensor, torch.Tensor)
    """
    radius = (-2 * input1.clamp(min=epsilon).log()).sqrt()
    angle = 2 * kPI * input2
    output1 = radius * angle.cos()
    output2 = radius * angle.sin()
    return output1, output2



from ._getter import get_feature
from ._getter import list_feature_names
from ._getter import list_features
from .container import FeatureList
from .container import ModuleOutput
from .features import Barrier
from .features import Empty
from .features import ExpiryTime  # deprecated
from .features import LogMoneyness
from .features import MaxLogMoneyness
from .features import MaxMoneyness
from .features import Moneyness
from .features import Ones
from .features import PrevHedge
from .features import Spot
from .features import TimeToMaturity
from .features import UnderlierSpot
from .features import Variance
from .features import Volatility
from .features import Zeros



from typing import Any
from typing import Callable

import torch
from torch import Tensor
from torch.testing import assert_close


def assert_monotone(
    fn: Callable[[Tensor], Tensor],
    x1: Tensor,
    x2: Tensor,
    increasing: bool = False,
    allow_equal: bool = False,
) -> None:
    """Assert ``fn`` is monotone function for :math:`x_1 > x_2`:

    .. math::
        f(x_1) > f(x_2)

    Args:
        fn (callable[[torch.Tensor], torch.Tensor]): Function to test.
        x1 (torch.Tensor): The first tensor.
        x2 (torch.Tensor): The second tensor.
        increasing (bool, default=True): Whether to test increasing monotonicity.
            Only False is supported.
        allow_equal (bool, default=False): Whether to allow :math:`f(x_1) = f(x_2)`.
            Only False is supported.

    Returns:
        None

    Raises:
        AssertionError: If fn is not monotone.
    """
    assert not increasing, "not supported"
    assert not allow_equal, "not supported"
    assert (x1 > x2).all(), "x1 > x2 must be satisfied"

    assert (fn(x1) < fn(x2)).all()


def assert_convex(
    fn: Callable[[Tensor], Tensor], x1: Tensor, x2: Tensor, alpha: float
) -> None:
    """Assert convexity.

    .. math::
        f(\\alpha * x_1 + (1 - \\alpha) * x_2) \\leq
        \\alpha * f(x_1) + (1 - \\alpha) * f(x_2)

    Args:
        fn (callable[[torch.Tensor], torch.Tensor]): Function to test.
            It should return a tensor with a single element.
        x1 (torch.Tensor): The first tensor.
        x2 (torch.Tensor): The second tensor.
        alpha (float): The parameter alpha.
    """
    y = fn(alpha * x1 + (1 - alpha) * x2)
    y1 = fn(x1)
    y2 = fn(x2)
    assert y <= alpha * y1 + (1 - alpha) * y2


def assert_cash_invariant(
    fn: Callable[[Tensor], Tensor], x: Tensor, c: float, **kwargs: Any
) -> None:
    """Assert cash invariance.

    .. math::
        f(x + c) = f(x) - c

    Args:
        fn (callable): Function to test cash invariance.
        x (torch.Tensor): The input tensor.
        c (float): The parameter c.
    """
    assert_close(fn(x + c), fn(x) - c, **kwargs)


def assert_cash_equivalent(
    fn: Callable[[Tensor], Tensor], x: Tensor, c: float, **kwargs: Any
) -> None:
    """Assert ``c`` is the cash equivalent of ``x``.
    ``fn(x) = fn(torch.full_like(x, c))``

    Args:
        fn (callable): Function to test cash equivalent.
        x (torch.Tensor): The input tensor.
        c (float): The parameter c.
        **kwargs: Keyword arguments to pass to ``assert_close``.
    """
    assert_close(fn(x), fn(torch.full_like(x, c)), **kwargs)



from collections import namedtuple
from typing import Optional
from typing import Tuple
from typing import Union
from typing import cast

import torch
from torch import Tensor

from pfhedge._utils.str import _addindent
from pfhedge._utils.typing import LocalVolatilityFunction
from pfhedge._utils.typing import TensorOrScalar

from ._utils import cast_state


class LocalVolatilityTuple(namedtuple("LocalVolatilityTuple", ["spot", "volatility"])):

    __module__ = "pfhedge.stochastic"

    def __repr__(self) -> str:
        items_str_list = []
        for field, tensor in self._asdict().items():

            items_str_list.append(field + "=\n" + str(tensor))
        items_str = _addindent("\n".join(items_str_list), 2)
        return self.__class__.__name__ + "(\n" + items_str + "\n)"

    @property
    def variance(self) -> Tensor:
        return self.volatility.square()


def generate_local_volatility_process(
    n_paths: int,
    n_steps: int,
    sigma_fn: LocalVolatilityFunction,
    init_state: Union[Tuple[TensorOrScalar, ...], TensorOrScalar] = (1.0,),
    dt: float = 1 / 250,
    dtype: Optional[torch.dtype] = None,
    device: Optional[torch.device] = None,
) -> LocalVolatilityTuple:
    r"""Returns time series following the local volatility model.

    The time evolution of the process is given by:

    .. math::
        dS(t) = \sigma_{\mathrm{LV}}(t, S(t)) dW(t) ,

    where :math:`\sigma_{\mathrm{LV}}` is the local volatility function.

    Args:
        n_paths (int): The number of simulated paths.
        n_steps (int): The number of time steps.
        init_state (tuple[torch.Tensor | float], default=(0.0,)): The initial state of
            the time series.
            This is specified by a tuple :math:`(S(0),)`.
            It also accepts a :class:`torch.Tensor` or a :class:`float`.
        sigma_fn (callable): The local volatility function.
            Its signature is ``sigma_fn(time: Tensor, spot: Tensor) -> Tensor``.
        dt (float, default=1/250): The intervals of the time steps.
        dtype (torch.dtype, optional): The desired data type of returned tensor.
            Default: If ``None``, uses a global default
            (see :func:`torch.set_default_tensor_type()`).
        device (torch.device, optional): The desired device of returned tensor.
            Default: If ``None``, uses the current device for the default tensor type
            (see :func:`torch.set_default_tensor_type()`).
            ``device`` will be the CPU for CPU tensor types and the current CUDA device
            for CUDA tensor types.

    Shape:
        - Output: :math:`(N, T)` where
          :math:`N` is the number of paths and
          :math:`T` is the number of time steps.

    Returns:
        (torch.Tensor, torch.Tensor): A namedtuple ``(spot, volatility)``.

    Examples:
        >>> from pfhedge.stochastic import generate_local_volatility_process
        ...
        >>> def sigma_fn(time: Tensor, spot: Tensor) -> Tensor:
        ...     a, b, sigma = 0.0001, 0.0004, 0.1000
        ...     sqrt_term = (spot.log().square() + sigma ** 2).sqrt()
        ...     return ((a + b * sqrt_term) / time.clamp(min=1/250)).sqrt()
        ...
        >>> _ = torch.manual_seed(42)
        >>> spot, volatility = generate_local_volatility_process(2, 5, sigma_fn)
        >>> spot
        tensor([[1.0000, 1.0040, 1.0055, 1.0075, 1.0091],
                [1.0000, 0.9978, 1.0239, 1.0184, 1.0216]])
        >>> volatility
        tensor([[0.1871, 0.1871, 0.1323, 0.1081, 0.0936],
                [0.1871, 0.1871, 0.1328, 0.1083, 0.0938]])
    """
    init_state = cast_state(init_state, dtype=dtype, device=device)

    spot = torch.empty(*(n_paths, n_steps), dtype=dtype, device=device)
    spot[:, 0] = init_state[0]
    volatility = torch.empty_like(spot)

    time = dt * torch.arange(n_steps).to(spot)
    dw = torch.randn_like(spot) * torch.as_tensor(dt).sqrt()

    for i_step in range(n_steps):
        sigma = sigma_fn(time[i_step], spot[:, i_step])
        volatility[:, i_step] = sigma
        if i_step != n_steps - 1:
            spot[:, i_step + 1] = spot[:, i_step] * (1 + sigma * dw[:, i_step])

    return LocalVolatilityTuple(spot, volatility)



from math import ceil
from typing import Optional
from typing import Tuple
from typing import cast

import torch
from torch import Tensor

from pfhedge._utils.doc import _set_attr_and_docstring
from pfhedge._utils.doc import _set_docstring
from pfhedge._utils.str import _format_float
from pfhedge._utils.typing import TensorOrScalar
from pfhedge.stochastic import generate_geometric_brownian

from .base import BasePrimary


class BrownianStock(BasePrimary):
    r"""A stock of which spot prices follow the geometric Brownian motion.

    .. seealso::
        - :func:`pfhedge.stochastic.generate_geometric_brownian`:
          The stochastic process.

    Args:
        sigma (float, default=0.2): The parameter :math:`\sigma`,
            which stands for the volatility of the spot price.
        mu (float, default=0.0): The parameter :math:`\mu`,
            which stands for the drift of the spot price.
        cost (float, default=0.0): The transaction cost rate.
        dt (float, default=1/250): The intervals of the time steps.
        dtype (torch.device, optional): Desired device of returned tensor.
            Default: If None, uses a global default
            (see :func:`torch.set_default_tensor_type()`).
        device (torch.device, optional): Desired device of returned tensor.
            Default: if None, uses the current device for the default tensor type
            (see :func:`torch.set_default_tensor_type()`).
            ``device`` will be the CPU for CPU tensor types and
            the current CUDA device for CUDA tensor types.

    Buffers:
        - spot (:class:`torch.Tensor`): The spot prices of the instrument.
          This attribute is set by a method :meth:`simulate()`.
          The shape is :math:`(N, T)` where
          :math:`N` is the number of simulated paths and
          :math:`T` is the number of time steps.

    Examples:
        >>> from pfhedge.instruments import BrownianStock
        >>>
        >>> _ = torch.manual_seed(42)
        >>> stock = BrownianStock()
        >>> stock.simulate(n_paths=2, time_horizon=5 / 250)
        >>> stock.spot
        tensor([[1.0000, 1.0016, 1.0044, 1.0073, 0.9930, 0.9906],
                [1.0000, 0.9919, 0.9976, 1.0009, 1.0076, 1.0179]])

        Using custom ``dtype`` and ``device``.

        >>> stock = BrownianStock()
        >>> stock.to(dtype=torch.float64, device="cuda:0")
        BrownianStock(..., dtype=torch.float64, device='cuda:0')
    """

    def __init__(
        self,
        sigma: float = 0.2,
        mu: float = 0.0,
        cost: float = 0.0,
        dt: float = 1 / 250,
        dtype: Optional[torch.dtype] = None,
        device: Optional[torch.device] = None,
    ) -> None:
        super().__init__()

        self.sigma = sigma
        self.mu = mu
        self.cost = cost
        self.dt = dt

        self.to(dtype=dtype, device=device)

    @property
    def default_init_state(self) -> Tuple[float, ...]:
        return (1.0,)

    @property
    def volatility(self) -> Tensor:
        """Returns the volatility of self.

        It is a tensor filled with ``self.sigma``.
        """
        return torch.full_like(self.get_buffer("spot"), self.sigma)

    @property
    def variance(self) -> Tensor:
        """Returns the volatility of self.

        It is a tensor filled with the square of ``self.sigma``.
        """
        return torch.full_like(self.get_buffer("spot"), self.sigma ** 2)

    def simulate(
        self,
        n_paths: int = 1,
        time_horizon: float = 20 / 250,
        init_state: Optional[Tuple[TensorOrScalar]] = None,
    ) -> None:
        """Simulate the spot price and add it as a buffer named ``spot``.

        The shape of the spot is :math:`(N, T)`, where :math:`N` is the number of
        simulated paths and :math:`T` is the number of time steps.
        The number of time steps is determinded from ``dt`` and ``time_horizon``.

        Args:
            n_paths (int, default=1): The number of paths to simulate.
            time_horizon (float, default=20/250): The period of time to simulate
                the price.
            init_state (tuple[torch.Tensor | float], optional): The initial state of
                the instrument.
                This is specified by a tuple :math:`(S(0),)` where
                :math:`S(0)` is the initial value of the spot price.
                If ``None`` (default), it uses the default value
                (See :attr:`default_init_state`).
                It also accepts a :class:`float` or a :class:`torch.Tensor`.

        Examples:
            >>> _ = torch.manual_seed(42)
            >>> stock = BrownianStock()
            >>> stock.simulate(n_paths=2, time_horizon=5 / 250, init_state=(2.0,))
            >>> stock.spot
            tensor([[2.0000, 2.0031, 2.0089, 2.0146, 1.9860, 1.9812],
                    [2.0000, 1.9838, 1.9952, 2.0018, 2.0153, 2.0358]])
        """
        if init_state is None:
            init_state = cast(Tuple[float], self.default_init_state)

        spot = generate_geometric_brownian(
            n_paths=n_paths,
            n_steps=ceil(time_horizon / self.dt + 1),
            init_state=init_state,
            sigma=self.sigma,
            mu=self.mu,
            dt=self.dt,
            dtype=self.dtype,
            device=self.device,
        )

        self.register_buffer("spot", spot)

    def extra_repr(self) -> str:
        params = ["sigma=" + _format_float(self.sigma)]
        if self.mu != 0.0:
            params.append("mu=" + _format_float(self.mu))
        if self.cost != 0.0:
            params.append("cost=" + _format_float(self.cost))
        params.append("dt=" + _format_float(self.dt))
        return ", ".join(params)


# Assign docstrings so they appear in Sphinx documentation
_set_docstring(BrownianStock, "default_init_state", BasePrimary.default_init_state)
_set_attr_and_docstring(BrownianStock, "to", BasePrimary.to)



from typing import List
from typing import Optional
from typing import Type

import torch
from torch import Tensor
from torch.nn import Module

from pfhedge._utils.str import _format_float
from pfhedge.instruments.derivative.base import BaseDerivative
from pfhedge.instruments.derivative.base import OptionMixin

from ._base import Feature
from ._base import StateIndependentFeature
from ._getter import FeatureFactory


# for mypy only
class OptionType(BaseDerivative, OptionMixin):
    pass


class Moneyness(StateIndependentFeature):
    """Moneyness of the derivative.

    Moneyness reads :math:`S / K` where
    :math:`S` is the spot price of the underlying instrument and
    :math:`K` is the strike of the derivative.

    Name:
        ``'moneyness'``

    Examples:
        >>> from pfhedge.features import Moneyness
        >>> from pfhedge.instruments import BrownianStock
        >>> from pfhedge.instruments import EuropeanOption
        ...
        >>> _ = torch.manual_seed(42)
        >>> derivative = EuropeanOption(BrownianStock(), maturity=5/250, strike=2.0)
        >>> derivative.simulate()
        >>> derivative.underlier.spot
        tensor([[1.0000, 1.0016, 1.0044, 1.0073, 0.9930, 0.9906]])
        >>> f = Moneyness().of(derivative)
        >>> f.get()
        tensor([[[0.5000],
                 [0.5008],
                 [0.5022],
                 [0.5036],
                 [0.4965],
                 [0.4953]]])
    """

    derivative: OptionType

    def __init__(self, log: bool = False) -> None:
        super().__init__()
        self.log = log

    def __str__(self) -> str:
        return "log_moneyness" if self.log else "moneyness"

    def get(self, time_step: Optional[int] = None) -> Tensor:
        return self.derivative.moneyness(time_step, log=self.log).unsqueeze(-1)


class LogMoneyness(Moneyness):
    r"""Log-moneyness of the derivative.

    Log-moneyness reads :math:`\log(S / K)` where
    :math:`S` is the spot price of the underlying instrument and
    :math:`K` is the strike of the derivative.

    Name:
        ``'log_moneyness'``
    """

    derivative: OptionType

    def __init__(self) -> None:
        super().__init__(log=True)


class TimeToMaturity(StateIndependentFeature):
    """Remaining time to the maturity of the derivative.

    Name:
        ``'time_to_maturity'``

    Examples:
        >>> from pfhedge.features import Moneyness
        >>> from pfhedge.instruments import BrownianStock
        >>> from pfhedge.instruments import EuropeanOption
        ...
        >>> _ = torch.manual_seed(42)
        >>> derivative = EuropeanOption(BrownianStock(), maturity=5/250, strike=2.0)
        >>> derivative.simulate()
        >>> f = TimeToMaturity().of(derivative)
        >>> f.get()
        tensor([[[0.0200],
                 [0.0160],
                 [0.0120],
                 [0.0080],
                 [0.0040],
                 [0.0000]]])
    """

    derivative: OptionType
    name = "time_to_maturity"

    def get(self, time_step: Optional[int] = None) -> Tensor:
        return self.derivative.time_to_maturity(time_step).unsqueeze(-1)


class ExpiryTime(TimeToMaturity):
    """Alias for ``TimeToMaturity``."""

    def __str__(self) -> str:
        return "expiry_time"


class UnderlierSpot(StateIndependentFeature):
    """Spot price of the underlier of the derivative.

    Name:
        ``'underlier_spot'``
    """

    def __init__(self, log: bool = False) -> None:
        super().__init__()
        self.log = log

    def __str__(self) -> str:
        return "underlier_log_spot" if self.log else "underlier_spot"

    def get(self, time_step: Optional[int] = None) -> Tensor:
        index = [time_step] if isinstance(time_step, int) else ...
        output = self.derivative.ul().spot[:, index].unsqueeze(-1)
        if self.log:
            output.log_()
        return output


class UnderlierLogSpot(UnderlierSpot):
    """Logarithm of the spot price of the underlier of the derivative.

    Name:
        ``'underlier_log_spot'``
    """

    def __init__(self):
        super().__init__(log=True)


class Spot(StateIndependentFeature):
    """Spot price of the derivative.

    Name:
        ``'spot'``
    """

    def __init__(self, log: bool = False) -> None:
        super().__init__()
        self.log = log

    def __str__(self) -> str:
        return "log_spot" if self.log else "spot"

    def get(self, time_step: Optional[int] = None) -> Tensor:
        index = [time_step] if isinstance(time_step, int) else ...
        output = self.derivative.spot[:, index].unsqueeze(-1)
        if self.log:
            output.log_()
        return output


class Volatility(StateIndependentFeature):
    """Volatility of the underlier of the derivative.

    Name:
        ``'volatility'``

    Examples:
    """

    name = "volatility"

    def get(self, time_step: Optional[int] = None) -> Tensor:
        index = [time_step] if isinstance(time_step, int) else ...
        return self.derivative.ul().volatility[:, index].unsqueeze(-1)


class Variance(StateIndependentFeature):
    """Variance of the underlier of the derivative.

    Name:
        ``'variance'``
    """

    name = "variance"

    def get(self, time_step: Optional[int]) -> Tensor:
        index = [time_step] if isinstance(time_step, int) else ...
        return self.derivative.ul().variance[:, index].unsqueeze(-1)


class PrevHedge(Feature):
    """Previous holding of underlier.

    Name:
        ``'prev_hedge'``
    """

    hedger: Module
    name = "prev_hedge"

    def get(self, time_step: Optional[int] = None) -> Tensor:
        if time_step is None:
            raise ValueError("time_step for prev_output should be specified")
        return self.hedger.get_buffer("prev_output")


class Barrier(StateIndependentFeature):
    """A feature which signifies whether the price of the underlier have reached
    the barrier. The output 1.0 means that the price have touched the barrier,
    and 0.0 otherwise.

    Args:
        threshold (float): The price level of the barrier.
        up (bool, default True): If ``True``, signifies whether the price has exceeded
            the barrier upward.
            If ``False``, signifies whether the price has exceeded the barrier downward.

    Examples:
        >>> from pfhedge.features import Barrier
        >>> from pfhedge.instruments import BrownianStock
        >>> from pfhedge.instruments import EuropeanOption
        ...
        >>> _ = torch.manual_seed(42)
        >>> derivative = EuropeanOption(BrownianStock(), maturity=5/250, strike=2.0)
        >>> derivative.simulate()
        >>> derivative.underlier.spot
        tensor([[1.0000, 1.0016, 1.0044, 1.0073, 0.9930, 0.9906]])
        >>> f = Barrier(threshold=1.004, up=True).of(derivative)
        >>> f.get()
        tensor([[[0.],
                 [0.],
                 [1.],
                 [1.],
                 [1.],
                 [1.]]])
    """

    def __init__(self, threshold: float, up: bool = True) -> None:
        super().__init__()
        self.threshold = threshold
        self.up = up

    def __repr__(self) -> str:
        params = [_format_float(self.threshold), "up=" + str(self.up)]
        return self._get_name() + "(" + ", ".join(params) + ")"

    def get(self, time_step: Optional[int] = None) -> Tensor:
        spot = self.derivative.ul().spot
        if time_step is None:
            if self.up:
                max = spot.cummax(-1).values
                output = (max >= self.threshold).to(spot.dtype)
            else:
                min = spot.cummin(-1).values
                output = (min <= self.threshold).to(spot.dtype)
        else:
            if self.up:
                max = spot[..., : time_step + 1].max(-1, keepdim=True).values
                output = (max >= self.threshold).to(spot.dtype)
            else:
                min = spot[..., : time_step + 1].min(-1, keepdim=True).values
                output = (min <= self.threshold).to(self.derivative.ul().spot.dtype)
        return output.unsqueeze(-1)


class Zeros(StateIndependentFeature):
    """A feature filled with the scalar value 0.

    Name:
        ``'zeros'``

    Examples:
        >>> from pfhedge.features import Zeros
        >>> from pfhedge.instruments import BrownianStock
        >>> from pfhedge.instruments import EuropeanOption
        ...
        >>> derivative = EuropeanOption(BrownianStock(), maturity=5/250, strike=2.0)
        >>> derivative.simulate()
        >>> f = Zeros().of(derivative)
        >>> f.get()
        tensor([[[0.],
                 [0.],
                 [0.],
                 [0.],
                 [0.],
                 [0.]]])
    """

    name = "zeros"

    def get(self, time_step: Optional[int] = None) -> Tensor:
        index = [time_step] if time_step is not None else ...
        return torch.zeros_like(self.derivative.ul().spot[..., index]).unsqueeze(-1)


class Ones(StateIndependentFeature):
    """A feature filled with the scalar value 1.

    Name:
        ``'ones'``

    Examples:
        >>> from pfhedge.features import Ones
        >>> from pfhedge.instruments import BrownianStock
        >>> from pfhedge.instruments import EuropeanOption
        ...
        >>> derivative = EuropeanOption(BrownianStock(), maturity=5/250, strike=2.0)
        >>> derivative.simulate()
        >>> f = Ones().of(derivative)
        >>> f.get()
        tensor([[[1.],
                 [1.],
                 [1.],
                 [1.],
                 [1.],
                 [1.]]])
    """

    name = "ones"

    def get(self, time_step: Optional[int] = None) -> Tensor:
        index = [time_step] if time_step is not None else ...
        return torch.ones_like(self.derivative.ul().spot[..., index]).unsqueeze(-1)


class Empty(StateIndependentFeature):
    """A feature filled with uninitialized data.

    Name:
        ``'empty'``

    Examples:
        >>> from pfhedge.features import Empty
        >>> from pfhedge.instruments import BrownianStock
        >>> from pfhedge.instruments import EuropeanOption
        ...
        >>> derivative = EuropeanOption(BrownianStock(), maturity=5/250, strike=2.0)
        >>> derivative.simulate()
        >>> f = Empty().of(derivative)
        >>> f.get()
        tensor([[[...],
                 [...],
                 [...],
                 [...],
                 [...],
                 [...]]])
    """

    name = "empty"

    def get(self, time_step: Optional[int] = None) -> Tensor:
        index = [time_step] if time_step is not None else ...
        return torch.empty_like(self.derivative.ul().spot[..., index]).unsqueeze(-1)


class MaxMoneyness(StateIndependentFeature):
    """Cumulative maximum of moneyness.

    Name:
        ``'max_moneyness'``

    Examples:
        >>> from pfhedge.features import MaxMoneyness
        >>> from pfhedge.instruments import BrownianStock
        >>> from pfhedge.instruments import EuropeanOption
        ...
        >>> _ = torch.manual_seed(42)
        >>> derivative = EuropeanOption(BrownianStock(), maturity=5/250, strike=2.0)
        >>> derivative.simulate()
        >>> derivative.underlier.spot
        tensor([[1.0000, 1.0016, 1.0044, 1.0073, 0.9930, 0.9906]])
        >>> f = MaxMoneyness().of(derivative)
        >>> f.get()
        tensor([[[0.5000],
                 [0.5008],
                 [0.5022],
                 [0.5036],
                 [0.5036],
                 [0.5036]]])
    """

    derivative: OptionType

    def __init__(self, log: bool = False) -> None:
        super().__init__()
        self.log = log

    def __str__(self) -> str:
        return "max_log_moneyness" if self.log else "max_moneyness"

    def get(self, time_step: Optional[int] = None) -> Tensor:
        return self.derivative.max_moneyness(time_step, log=self.log).unsqueeze(-1)


class MaxLogMoneyness(MaxMoneyness):
    """Cumulative maximum of log Moneyness.

    Name:
        ``'max_log_moneyness'``
    """

    derivative: OptionType

    def __init__(self) -> None:
        super().__init__(log=True)


FEATURES: List[Type[Feature]] = [
    Empty,
    ExpiryTime,
    TimeToMaturity,
    LogMoneyness,
    MaxLogMoneyness,
    MaxMoneyness,
    Moneyness,
    PrevHedge,
    Variance,
    Volatility,
    Zeros,
    Spot,
    UnderlierSpot,
]

for cls in FEATURES:
    FeatureFactory().register_feature(str(cls()), cls)



from math import floor

import torch
from torch import Tensor

from pfhedge._utils.doc import _set_attr_and_docstring
from pfhedge._utils.doc import _set_docstring
from pfhedge._utils.str import _format_float
from pfhedge.nn.functional import european_forward_start_payoff

from ..primary.base import BasePrimary
from .base import BaseDerivative


class EuropeanForwardStartOption(BaseDerivative):
    r"""European forward start option.

    The payoff is given by

    .. math ::
        \mathrm{payoff} = \max(S_T / S_{T'} - K, 0) ,

    where
    :math:`S_T` is the underlier's spot price at maturity,
    :math:`S_{T'}` is the underlier's spot price at ``start``, and
    :math:`K` is the strike.

    Note:
        If ``start`` is not divisible by the interval of the time step,
        it is rounded off so that the start time is the largest value
        that is divisible by the interval and less than ``start``.

    .. seealso::
        - :func:`pfhedge.nn.functional.european_forward_start_payoff`

    Args:
        underlier (:class:`BasePrimary`): The underlying instrument of the option.
        strike (float, default=1.0): The strike value of the option.
        maturity (float, default=20/250): The maturity of the option.
        start (float, default=10/250): The start of the option.

    Attributes:
        dtype (torch.dtype): The dtype with which the simulated time-series are
            represented.
        device (torch.device): The device where the simulated time-series are.

    Examples:
        >>> import torch
        >>> from pfhedge.instruments import BrownianStock
        >>> from pfhedge.instruments import EuropeanForwardStartOption
        >>>
        >>> _ = torch.manual_seed(42)
        >>> derivative = EuropeanForwardStartOption(BrownianStock(), maturity=5/250, start=2/250)
        >>> derivative.simulate(n_paths=2)
        >>> derivative.underlier.spot
        tensor([[1.0000, 1.0016, 1.0044, 1.0073, 0.9930, 0.9906],
                [1.0000, 0.9919, 0.9976, 1.0009, 1.0076, 1.0179]])
        >>> derivative.underlier.spot[:, -1] / derivative.underlier.spot[:, 2]
        tensor([0.9862, 1.0203])
        >>> derivative.payoff()
        tensor([0.0000, 0.0203])
    """

    def __init__(
        self,
        underlier: BasePrimary,
        strike: float = 1.0,
        maturity: float = 20 / 250,
        start: float = 10 / 250,
    ) -> None:
        super().__init__()
        self.register_underlier("underlier", underlier)
        self.strike = strike
        self.maturity = maturity
        self.start = start

    def extra_repr(self) -> str:
        params = [
            "strike=" + _format_float(self.strike),
            "maturity=" + _format_float(self.maturity),
            "start=" + _format_float(self.start),
        ]
        return ", ".join(params)

    def _start_index(self) -> int:
        return floor(self.start / self.ul().dt)

    def payoff_fn(self) -> Tensor:
        return european_forward_start_payoff(
            self.ul().spot, strike=self.strike, start_index=self._start_index()
        )


# Assign docstrings so they appear in Sphinx documentation
_set_attr_and_docstring(EuropeanForwardStartOption, "simulate", BaseDerivative.simulate)
_set_attr_and_docstring(EuropeanForwardStartOption, "to", BaseDerivative.to)
_set_attr_and_docstring(EuropeanForwardStartOption, "ul", BaseDerivative.ul)
_set_attr_and_docstring(EuropeanForwardStartOption, "list", BaseDerivative.list)
_set_docstring(EuropeanForwardStartOption, "payoff", BaseDerivative.payoff)



from typing import Optional

import torch
from torch import Tensor

from pfhedge._utils.doc import _set_attr_and_docstring
from pfhedge._utils.doc import _set_docstring
from pfhedge._utils.str import _format_float
from pfhedge.nn.functional import european_binary_payoff

from ..primary.base import BasePrimary
from .base import BaseDerivative
from .base import OptionMixin


class EuropeanBinaryOption(BaseDerivative, OptionMixin):
    r"""European binary option.

    The payoff of a European binary call option is given by

    .. math::
        \mathrm{payoff} =
        \begin{cases}
            1 & (S \geq K) \\
            0 & (\text{otherwise})
        \end{cases} ,

    where
    :math:`S` is the underlier's spot price at maturity and
    :math:`K` is the strike.

    The payoff of a European binary put option is given by

    .. math::
        \mathrm{payoff} =
        \begin{cases}
            1 & (S \leq K) \\
            0 & (\text{otherwise})
        \end{cases} .

    .. seealso::
        - :func:`pfhedge.nn.functional.european_binary_payoff`

    Args:
        underlier (:class:`BasePrimary`): The underlying instrument of the option.
        call (bool, default=True): Specifies whether the option is call or put.
        strike (float, default=1): The strike price of the option.
        maturity (float, default=20/250): The maturity of the option.

    Attributes:
        dtype (torch.dtype): The dtype with which the simulated time-series are
            represented.
        device (torch.device): The device where the simulated time-series are.

    Examples:
        >>> import torch
        >>> from pfhedge.instruments import BrownianStock
        >>> from pfhedge.instruments import EuropeanBinaryOption
        >>>
        >>> _ = torch.manual_seed(42)
        >>> derivative = EuropeanBinaryOption(BrownianStock(), maturity=5/250)
        >>> derivative.simulate(n_paths=2)
        >>> derivative.underlier.spot
        tensor([[1.0000, 1.0016, 1.0044, 1.0073, 0.9930, 0.9906],
                [1.0000, 0.9919, 0.9976, 1.0009, 1.0076, 1.0179]])
        >>> derivative.payoff()
        tensor([0., 1.])
    """

    def __init__(
        self,
        underlier: BasePrimary,
        call: bool = True,
        strike: float = 1.0,
        maturity: float = 20 / 250,
        dtype: Optional[torch.dtype] = None,
        device: Optional[torch.device] = None,
    ) -> None:
        super().__init__()
        self.register_underlier("underlier", underlier)
        self.call = call
        self.strike = strike
        self.maturity = maturity

        # TODO(simaki): Remove later. Deprecated for > v0.12.3
        if dtype is not None or device is not None:
            self.to(dtype=dtype, device=device)
            raise DeprecationWarning(
                "Specifying device and dtype when constructing a Derivative is deprecated."
                "Specify them in the constructor of the underlier instead."
            )

    def extra_repr(self) -> str:
        params = []
        if not self.call:
            params.append("call=" + str(self.call))
        params.append("strike=" + _format_float(self.strike))
        params.append("maturity=" + _format_float(self.maturity))
        return ", ".join(params)

    def payoff_fn(self) -> Tensor:
        return european_binary_payoff(
            self.ul().spot, call=self.call, strike=self.strike
        )


# Assign docstrings so they appear in Sphinx documentation
_set_attr_and_docstring(EuropeanBinaryOption, "simulate", BaseDerivative.simulate)
_set_attr_and_docstring(EuropeanBinaryOption, "to", BaseDerivative.to)
_set_attr_and_docstring(EuropeanBinaryOption, "ul", BaseDerivative.ul)
_set_attr_and_docstring(EuropeanBinaryOption, "list", BaseDerivative.list)
_set_docstring(EuropeanBinaryOption, "payoff", BaseDerivative.payoff)
_set_attr_and_docstring(EuropeanBinaryOption, "moneyness", OptionMixin.moneyness)
_set_attr_and_docstring(
    EuropeanBinaryOption, "log_moneyness", OptionMixin.log_moneyness
)
_set_attr_and_docstring(
    EuropeanBinaryOption, "time_to_maturity", OptionMixin.time_to_maturity
)



from .base import BaseInstrument
from .base import Instrument
from .derivative.american_binary import AmericanBinaryOption
from .derivative.base import BaseDerivative
from .derivative.base import BaseOption
from .derivative.base import Derivative
from .derivative.base import OptionMixin
from .derivative.cliquet import EuropeanForwardStartOption
from .derivative.european import EuropeanOption
from .derivative.european_binary import EuropeanBinaryOption
from .derivative.lookback import LookbackOption
from .derivative.variance_swap import VarianceSwap
from .primary.base import BasePrimary
from .primary.base import Primary
from .primary.brownian import BrownianStock
from .primary.cir import CIRRate
from .primary.heston import HestonStock
from .primary.local_volatility import LocalVolatilityStock
from .primary.vasicek import VasicekRate
